/*
   program : various

   $Source: /home/yee/C/RCS/util.c,v $
   $Author: yee $
   $Date: 1999/03/09 01:52:38 $
   $Revision: 1.5 $

   Misc. utility functions.

   Compile Options : none

   Runtime Parameters : none

   Assumptions : none known

   Bugs : none known

   To Do :
     1) figure out cleaner logic for StringCopy() and StringLength().
     2) separate out buffered input from line parsing in fgetline().

   $Log: util.c,v $
   Revision 1.5  1999/03/09 01:52:38  yee
   *** empty log message ***

   Revision 1.3  1999/02/13 07:21:15  yee
   altered fgetline() so that the character input function can vary.
   Needed to support curses input as well as normal stdin input.

   Copyright 1999 by Conway Yee.  All rights reserved.
 */
char util[] = "$Id: util.c,v 1.5 1999/03/09 01:52:38 yee Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "style.h"
#include "util.h"

typedef struct skeleton{    /* used by AddElement(). */
  struct skeleton *next;
  struct skeleton *prev;
} skeleton;

PRIVATE int chbuff
(
 void *stream,   /* input pointer */
 int (*func)(),  /* function to do inputting */
 int ch          /* char to push. NULL if pull char */
);

/*****************************************************************************/
/*****************************************************************************/
 
/* this functions serves as a storage buffer for fgetline(). */

PUBLIC char *strbuff
(
 FILE *fptr,      /* file to be read */
 char *str        /* string to push. NULL if pull string */
)
{
  static char *buf=NULL;     /* buffered string */
  char *p=NULL;              /* string to return */

  if(str)
    buf=str;  /* push string to buffer */
  else
    if(buf) { /* pull string */
      p=buf;
      buf=NULL;
    } else
      p=fgetline(fptr,fgetc);

  return(p);
}

/*****************************************************************************/

/* this function reads and returns a line from the data file.  If EOF
   is reached, NULL is returned.

   Array allocation is handled internally since the length of a line
   can not be determined a priori.

   When AN EOF is reached (any file), the space for the data line is
   freed.  Since there shouldn't be too many input files, this
   restriction isn't very serious. */

#define QUANTUM 1024
PUBLIC char *fgetline
(
 void *stream,                   /* input pointer */
 int (*func)()                   /* function to read individual character*/
)
{
  static char *str=NULL;      /* input data line read */
  static int qnum=1;          /* number of quanta allocated */
  int space=qnum*QUANTUM-1;   /* space left in allocated string */
  int ch;                     /* character read from fptr */
  int cur=0;                  /* current string position */
  BOOLEAN inquote=FALSE;      /* is text within quotes? */

  if(str==NULL) str=malloc(QUANTUM*sizeof(char));

  do {                                /* get rid of initial white space */
    ch=chbuff(stream,func,NUL);
  } while(isspace(ch));

  if(ch!=EOF) {
    chbuff(stream,func,ch);

    do {
      if(space==0) {                     /* make sure space available in str */
        qnum++;
        str=realloc(str,qnum*QUANTUM*sizeof(char));
        space=QUANTUM-1;
      }

      ch=chbuff(stream,func,NUL);

      if(ch==BACKSLASH) {       /* line continue character */
        do {
          ch=chbuff(stream,func,NUL);
        } while(ch!=NL);
        ch=chbuff(stream,func,NUL);
      }

      if(ch==QUOTE) inquote = !inquote; /* test if in string */

      if((isspace(ch)||(ch==EOF))&&((cur==0)||isspace(str[cur-1]))&&(!inquote))
         continue;   /* two white spaces in a row */

      if((ch!=NL)&&isspace(ch)) ch=SPACE;

      switch(ch) {        /* put ch into str */
      case NL:
      case EOF:
      case POUND:
        str[cur++]=NUL;
        break;
      default:
        space--;
        str[cur++]=ch;
      }

      if(ch==POUND) {                /* eliminate comment */
        do {
          ch=chbuff(stream,func,NUL);
        } while((ch!=NL)&&(ch!=EOF));
      }  
    } while((ch!=NL)&&(ch!=EOF));
  }

  if(ch==EOF) {        /* DONE!! */
    free(str);
    str=NULL;
  }

  return(str);
}
#undef QUANTUM

/*****************************************************************************/

/* returns length of string delimitted by tok */

PUBLIC int StringLength
(
 char *s,         /* data string to be read */
 char tok         /* delimiting token */
)
{
  int i;  /* counter */

  if(isspace(tok)) {
    for(i=1;(isspace(*s))&&(*s!=NUL);s++);  /* eliminate initial white space */
    for(;(*s!=NUL)&&(!isspace(*s));s++,i++);    /* count */
    i--;
  } else {
    for(i=0;(*s!=tok)&&(*s!=NUL);s++);    /* get rid of initial garbage */
    s++;
    for(;(*s!=tok)&&(*s!=NUL);s++,i++);
  }

  return(i);
}

/*****************************************************************************/

/* copys a string delimitted by tok from s to ct. It returns the character
   past the end delimiter */

PUBLIC char *StringCopy
(
 char *d,      /* destination string */
 char *s,      /* source string */
 char tok,     /* token */
 int n         /* max number of characters to search */
)
{
  if(isspace(tok)) {
                                            /* eliminate initial white space */
    for(;(n>0)&&(isspace(*s)&&(*s!=NUL));s++,n--);

    for(;(n>0)&&(s!=NUL)&&!isspace(*s);d++,s++,n--)  /* copy */
      *d = *s;

  } else {
    for(;(*s!=tok)&&(n>0)&&(*s!=NUL);s++,n--);  /* find & pass start token */
    s++;

    for(;(*s!=tok)&&(n>0);d++,s++,n--)  /* copy string */
      *d = *s;

    s++;
    *d=NUL;
  }
                                                /* elminate end white space */
  for(;(*s!=tok)&&isspace(*s)&&(*s!=NUL)&&(n>0);s++,n--);
  return(s);
}

/*****************************************************************************/

/* this is a generic function to add elements to a double linked list.

   NOTE: the elements must have next and prev as their first two entries. */

PUBLIC void *AddElement
(
 void *list,  /* pointer to existing double linked list */
 int size     /* size of element to add */
)
{
  skeleton *t=calloc(1,size);       /* temp pointer to new element */

  if(list) {                        /* add to existing ring */
    t->next=((skeleton *)list);
    t->prev=((skeleton *)list)->prev;
    ((skeleton *)list)->prev=t;
    t->prev->next=t;
  } else {                             /* first node */
    t->prev=t->next=t;
    list=t;
  }

  return (list);
}

/*****************************************************************************/

/* this function serves as an internal character storage buffer for
   fgetline(). */

PRIVATE int chbuff
(
 void *stream,   /* input pointer */
 int (*func)(),  /* function to do inputting */
 int ch          /* char to push. NULL if pull char */
)
{
  static char buf=NUL;       /* buffered string */
  char p=NUL;                /* char to return */

  if(ch!=NUL)
    buf=ch;  /* push char to buffer */
  else
    if(buf) { /* pull char */
      p=buf;
      buf=NUL;
    } else
      p=func(stream);

  return(p);
}

